import abc
from abc import ABC
from functools import cache
from typing import Generic, Optional, Union, Tuple, Any, Dict, Set, TypeVar

from sakkara.model.base import ModelComponent
from sakkara.model.math_op import MathOpBase
from sakkara.relation.groupset import GroupSet

from sakkara.relation.representation import Representation

S = TypeVar('S', bound=Any)
T = TypeVar('T', bound=ModelComponent)


class Composable(MathOpBase, ABC, Generic[S, T]):
    """
    Base class for a component that can be built with underlying subcomponents



    :param name: Name of the corresponding variable to register in PyMC.

    :param group: Group of which the component is defined for.

    :param members: Subset of members of column the component is defined for.

    :param subcomponents: Dict of underlying ModelComponent objects.

    """

    def __init__(self, name: Optional[str], group: Optional[Union[str, Tuple[str, ...]]], subcomponents: Dict[S, T]):
        super().__init__()
        self.name = name
        if isinstance(group, str):
            self.group = (group,)
        elif group is None:
            self.group = tuple()
        else:
            self.group = group
        self.subcomponents = subcomponents
        self.base_representation = None
        self.components_representation = None

    @abc.abstractmethod
    def __getitem__(self, item: Any) -> T:
        raise NotImplementedError

    def get_name(self) -> Optional[str]:
        return self.name

    def set_name(self, name: str) -> None:
        self.name = name

    def build_components(self, groupset: GroupSet) -> None:
        for param_name, component in self.subcomponents.items():
            if component.get_name() is None:
                component.set_name(f'{param_name}_{self.get_name()}')
            if component.variable is None:
                component.build(groupset)

    def build_representation(self, groupset: GroupSet):
        self.representation = Representation()
        self.base_representation = Representation()
        self.components_representation = Representation()

        for g in self.group:
            self.base_representation.add_group(groupset[g])

        for component in self.subcomponents.values():
            self.components_representation = self.components_representation.merge(component.representation)

        self.representation = self.base_representation.merge(self.components_representation)

        if len(self.representation.groups) == 0:
            self.representation.add_group(groupset['global'])

    def clear(self):
        self.variable = None
        self.representation = None
        for c in self.subcomponents.values():
            c.clear()

    @cache
    def retrieve_groups(self) -> Set[str]:
        group = set()
        for k, v in self.subcomponents.items():
            parent_groups = v.retrieve_groups()
            group = group.union(parent_groups)
        if self.group is not None:
            group = group.union(self.group)

        return group

    def dims(self):
        return tuple(map(str, self.representation.groups))



from dataclasses import dataclass
from typing import Dict

import numpy as np
import pandas as pd

from sakkara.relation.group import Group


@dataclass(frozen=True)
class GroupSet:
    """
    Set of group nodes used for PyMC model creation
    """
    groups: Dict[str, Group]

    def __getitem__(self, item: str):
        return self.groups[item]

    def coords(self) -> Dict[str, np.ndarray]:
        coords_dict = {}
        for k, v in self.groups.items():
            coords_dict[k] = v.members
        return coords_dict


def get_parent_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Create a matrix of parent mappings between columns in a dataframe

    :param df: Dataframe (with C columns) to base parent mappings of.

    :return Dataframe of size CxC with parent mappings. Rows are sorted from highest to lowest hierarchy level.
    Element (i,j) True means that j is a parent to i.
    """
    groups = list(df.columns)
    n_uniques = df.nunique(axis=0)
    counts_df = pd.DataFrame(index=groups, columns=groups, data=False)
    for i in range(len(groups)):
        # Parent to child is when child value always gives parent value on same row, and there are more unique values of
        # child
        counts_df.loc[groups[i], groups[:i] + groups[i + 1:]] = np.logical_and(
            df.groupby(groups[i]).nunique().max() == 1,
            n_uniques[groups[i]] > n_uniques[groups[:i] + groups[i + 1:]])

    # Sort counts_df from the lowest child in the first row to the highest parent in the last row
    counts_df['rank'] = counts_df.sum(axis=1)
    counts_df.sort_values(by='rank', inplace=True, ascending=True)
    return counts_df.loc[:, df.columns]


def get_twin_df(df):
    """
    Create a matrix of twin relations between columns in a dataframe

    :param df: Dataframe (with C columns) to base twin mappings of.

    :return Dataframe of size CxC with twin mappings.
    """
    groups = list(df.columns)
    counts_df = pd.DataFrame(index=groups, columns=groups, data=False)
    # Twin is when both group a value always gives group b value on corresponding row, and vice versa
    for i in range(len(groups)):
        counts_df.loc[groups[i], groups[:i] + groups[i + 1:]] = df.groupby(groups[i]).nunique().max() == 1
    return np.logical_and(counts_df, counts_df.T)


def init(df: pd.DataFrame) -> GroupSet:
    """
    Init a group set from a dataframe

    :param df: DataFrame containing only the group columns
    :return: GroupSet created from the input DataFrame
    """

    groups = {column: Group(column, df[column].unique()) for column in df.columns}

    parent_df = get_parent_df(df)
    for i, (group_name, is_parent) in enumerate(parent_df.iterrows()):
        # Create data frame with only rows with first appearance of each unique value of group column
        map_df = df.set_index(group_name)
        map_df = map_df.loc[~map_df.index.duplicated(keep='first'), :]
        for parent_name in df.columns[is_parent]:
            parent_mapping = groups[parent_name].mapping.loc[map_df[parent_name], parent_name]
            groups[group_name].add_parent(groups[parent_name], parent_mapping.values)
            groups[parent_name].add_child(groups[group_name])

    twin_df = get_twin_df(df)
    for group_name, is_twin in twin_df.iterrows():
        for twin_name in df.columns[is_twin]:
            groups[group_name].add_twin(groups[twin_name])

    return GroupSet(groups)



from abc import ABC
from typing import Tuple, Union, Any, Dict

import pymc as pm
from pytensor import tensor as pt
import numpy as np
from numpy import typing as npt

from sakkara.model.base import ModelComponent
from sakkara.model.composable.base import Composable, T
from sakkara.model.fixed.base import FixedValueComponent
from sakkara.relation.groupset import GroupSet


class GroupComponent(Composable[Tuple[str, ...], T], ABC):
    """
    Class for specifying components for each member of a group individually

    :param group: Group of which the component is defined for.
    :param membercomponents: Dictionary with key indicating member (corresponding to DataFrame value) and value its
            corresponding value (ModelComponent or other)
    :param name: Name of the corresponding variable to register in PyMC.
    """

    def __init__(self, group: Union[str, Tuple[str, ...]], name: str = None, membercomponents: Dict[Any, Any] = None):
        super().__init__(name, group, dict())
        if membercomponents is not None:
            for k, v in membercomponents.items():
                self.add(k, v)

    def __getitem__(self, item: Any):
        return self.subcomponents[item if isinstance(item, tuple) else (item,)]

    def add(self, member: Any, component: Union[float, npt.NDArray, ModelComponent]) -> None:
        """
        Add component for a member to the GroupComponent

        :param member: Key for the group member, corresponding to Dataframe value.
        :param component: Value or component of the given member.
        """
        if not isinstance(component, ModelComponent):
            component = FixedValueComponent(component)
        key = member if isinstance(member, tuple) else (member,)
        self.subcomponents[key] = component

    def prebuild(self, groupset: GroupSet) -> None:
        self.build_components(groupset)

    def build_variable(self) -> None:
        base_members = self.base_representation.get_members()

        # Get all the member tuples, raveled
        if len(base_members) == 1:
            all_member_tuples = [(x,) for x in np.ravel(base_members[0])]
        else:
            all_member_tuples = [(x,) for x in zip(map(np.ravel, base_members))]

        if not all(m in self.subcomponents for m in all_member_tuples):
            raise ValueError('All member of group component must be specified.')

        # Get the member array for each group, re-order with mapping and flatten
        base_to_repr_map = self.base_representation.map_to(self.representation)
        base_members_reordered = tuple(map(lambda x: x[base_to_repr_map].ravel(), base_members))

        # Create array for adding
        member_tensor = np.empty(len(all_member_tuples), dtype=object)

        for i, member_tuple in enumerate(all_member_tuples):
            # Get the component
            component = self.subcomponents[member_tuple]
            # Get the variable, re-order to the representation and flatten
            member_variable = component.variable[component.representation.map_to(self.representation)].ravel()

            # Mask all the entries of the group variable (with self.representation) corresponding to this member
            mask = np.ones(np.prod(self.representation.get_shape()), dtype=bool)
            for j, m in enumerate(member_tuple):
                mask *= base_members_reordered[j] == m

            member_tensor[i] = member_variable[mask]

        # Stack the member-wise variables into the group variable
        # NOTE! We assume that self.base_representation correspond to the first groups of self.representation
        # This might not be the case if a child of a group in self.base_representation is included in
        # self.components_representation (Hence, this is not supported)
        full_tensor = pt.stack(member_tensor.tolist()).ravel().reshape(self.representation.get_shape())

        # Create the group variable, wrapped with Deterministic
        self.variable = pm.Deterministic(name=self.name, var=full_tensor, dims=self.dims())



from typing import Set, Any, Tuple, Callable, Union, List

import numpy as np
import numpy.typing as npt
import pandas as pd


class Group:
    """
    Class corresponding to a data frame column, with capabilities of handling relations to other groups.

    :param name: Name of the group, should be same as the column
    :param members: The names of the unique column values, the members, ordered by their first appearance.
    """
    def __init__(self, name: str, members: npt.NDArray[Any]):
        self.name = name
        self.members = members
        self.parents = set()
        self.children = set()
        self.twins = {self}
        self.mapping = pd.DataFrame(index=members, data={name: np.arange(len(members))})

    def add_child(self, child: 'Group') -> None:
        """
        Add a child relation to this group

        :param child: The child group
        """
        self.children.add(child)

    def add_parent(self, parent: 'Group', parent_mapping: npt.NDArray[int]) -> None:
        """
        Add a parent relation to this group

        :param parent: The parent group
        :param parent_mapping: Indices of parent members (of the parent group), ordered by the corresponding member of
            this group
        """
        self.parents.add(parent)
        self.mapping[parent.name] = parent_mapping

    def add_twin(self, twin: 'Group') -> None:
        """
        Add a twin relation to this group

        :param twin: The twin group
        """
        self.twins.add(twin)
        self.mapping[twin.name] = np.arange(len(self.mapping))

    def __str__(self):
        return self.name

    def __len__(self):
        return len(self.mapping)

    def __lt__(self, other) -> bool:
        return str(self) < str(other)



from typing import Union, Tuple, Any

import numpy as np
import numpy.typing as npt

from sakkara.relation.group import Group


class Representation:
    """
    Class for a tuple of groups, reduced to its minimal representation. E.g., a parent can be represented by its child,
    hence when creating a `Representation` object of the two only the child will be added to the tuple.

    :param \*groups: :class:`Group` objects to include in representation
    """
    def __init__(self, *groups: 'Group'):
        self.groups = []
        for g in groups:
            self.add_group(g)

    def add_group(self, group: Group) -> None:
        """
        Add group to the representation. It will not be added if there already is a twin or a child in
        the representation. If there is a parent to the group in the representation, the group will be added and the
        parent will be removed.
        """
        # Check that group is not already represented by twin or child
        representatives = group.twins.union(group.children)
        if any(map(lambda t: t in representatives, self.groups)):
            return

        # Add group to representation
        self.groups.append(group)

        # Check if newly added group's parents are represented in group
        for parent in group.parents:
            # Remove if they are included (due to redundancy)
            if parent in self.groups:
                self.groups.remove(parent)

    def merge(self, other: 'Representation') -> 'Representation':
        """
        Create a new `Representation` by adding upp groups from this object and from another.

        :param other: Other `Representation` to merge with

        :return: Merged representation. Note that all groups of the merged has either a twin or a parent group in at
            least one of the two input representations.
        """
        merged = Representation(*self.groups)
        for group in other.groups:
            merged.add_group(group)

        return merged

    def is_mappable_to(self, target: 'Representation') -> bool:
        """
        Check if representation is mappable to another representation, i.e., groups of the other representation have
        twins or parents among the groups of this representation.

        :return: `True` is self is mappable to target
        """
        target_represents = set()
        for target_group in target.groups:
            target_represents = target_represents.union(target_group.twins).union(target_group.parents)
        return all(map(lambda t: t in target_represents, self.groups))

    def get_shape(self) -> Tuple[int, ...]:
        """
        Get the shape of a variable corresponding to a component with this representation
        """
        return tuple(map(len, self.groups))

    def __eq__(self, other: 'Representation'):
        return len(self.groups) == len(other.groups) and all(map(lambda x, y: x == y, self.groups, other.groups))

    def get_member_array(self, group: Group) -> npt.NDArray[Any]:
        """
        Get an array shaped as a variable with this representation, where each value correspond to the member of the
        i:th group.
        """
        group_index = self.groups.index(group)
        # Get shape of representation, but target_group excluded
        other_shape = self.get_shape()[:group_index] + self.get_shape()[group_index + 1:]
        # Create flat array or target_group members repeated, as if i == 0
        member_array = np.repeat(group.members, np.prod(other_shape))
        # Reshape target_members to have the same order as target's representation
        member_array = member_array.reshape((len(group),) + other_shape)
        # Swap i:th and 0:th axes to handle the case when i != 0
        member_array = np.swapaxes(member_array, 0, group_index)

        return member_array

    def map_to(self, target: 'Representation') -> Union[Tuple[npt.NDArray[int], ...], slice]:
        """
        Create indexing for a variable of this representation to another target representation.

        :param target: The representation that corresponds to the shape the variable should be in. Each group of self
            representation must have a twin or a child in the target representation.

        :return: Tuple of index arrays, or slice object, to be used for indexing on array object. For more information
            about the format of this tuple, see https://numpy.org/doc/stable/user/basics.indexing.html#advanced-indexing.

        """
        if not self.is_mappable_to(target):
            raise ValueError(
                'Representation cannot be mapped to target. Instead do mapping to a representation of this merged with target.')

        mapping = []

        for group in self.groups:
            for i, target_group in enumerate(target.groups):
                # Match any group of target representation that is either child or twin to this
                representatives = {group}.union(group.children).union(group.twins)

                if target_group in representatives:
                    # Get the array of members of target group, shaped according to target representation
                    target_members = target.get_member_array(target_group)
                    # Fetch mapping from target_members into members of group
                    this_group_members = target_group.mapping.loc[target_members.ravel(), group.name].values
                    # Append mapping, reshaped to target representation's shape
                    mapping.append(this_group_members.reshape(target.get_shape()))

                    # Finding more representatives not needed
                    break

        return tuple(mapping)

    def get_members(self) -> Tuple[npt.NDArray[Any], ...]:
        """
        Get array of members corresponding to the indices of a variable with the representation

        :return: One array per group of representation filled with corresponding member names, so that the i:th array
            correspond to the i:th group of the representation
        """
        return tuple(map(lambda group: self.get_member_array(group), self.groups))
